/***********************************************************

    hbs reduction subroutine: hbstfx.c

                                   1998/01/24  K. Kawabata

     transformation of image in order to correct distortions
                       (using shift x-direction)


    usage: % hbstfy infile ref_table outfile

      infile -- input image file (FITS format)

      ref_table -- fitted coeficients table (made by hbs1to1x.cl)

      outfile -- output spectrum file (FITS format)

************************************************************/
		  
#include<stdio.h>
#include<string.h>
#include<stdlib.h>
#include<math.h>
#include"fitsio.h"

#include"hbsdefine.h"

#define FITORDER 4 /* fitting order in hbs1to1y.cl */
#define LMDORDER 8 /* fitting order for x(wavelength)-direction */
/* These orders are the number of terms in fitting polynomial */

int readpeakfit( char *infile, double coef[][FITORDER]  ){

  FILE *fp;
  int n=0;
  char buf[100], last_buf[100]="";

  if( ( fp = fopen( infile, "rt" ) ) == NULL ){
    return( -1 );
  }

  while( feof( fp ) == 0 ){
    fgets( buf, 100, fp );
    if( strcmp( buf, last_buf ) == 0 )
      continue;
    sscanf( buf, "%lf %lf %lf %lf", &coef[n][0], &coef[n][1],\
	    &coef[n][2], &coef[n][3] );
    n++;
    strcpy( last_buf, buf );
  }
  fclose( fp );
  return( n );
}

int readlmd( char *infile, double lmd[]  ){

  FILE *fp;
  int n=0;
  char buf[100], last_buf[100]="";
  double a;

  if( ( fp = fopen( infile, "rt" ) ) == NULL ){
    return( -1 );
  }

  while( feof( fp ) == 0 ){
    fgets( buf, 100, fp );
    if( strcmp( buf, last_buf ) == 0 )
      continue;
    sscanf( buf, "%lf %lf", &a, &lmd[n] );
    n++;
    strcpy( last_buf, buf );
  }
  fclose( fp );
  return( n );
}

int scanfile( char *infile, float count[][NX1]  ){

  fitsfile *fp;
  int i, anynul, status;
  char buf[100];
  float nulval = 0.;
  long naxis1, naxis2, npix;
  char comment[100];

  status = 0;

  if( fits_open_file( &fp, infile, READONLY, &status ) != 0 ){
    printf( "cannot open file '%s' status=%d\n", infile, status );
    return( status );
  }
  if( fits_read_key( fp, TLONG, "NAXIS1", &naxis1,\
      comment, &status ) != 0 ){
    return( status );
  }
  if( fits_read_key( fp, TLONG, "NAXIS2", &naxis2,\
      comment, &status ) != 0 ){
    return( status );
  }
  if( naxis1 != NX1 || naxis2 != NY1 ){
    printf( "error: definition of NX1 or NY1 in 'hbsdefine.h'\n" );
    return( -1 );
  }
  npix = naxis1 * naxis2;

  if( fits_read_img( fp, TFLOAT, 1L, npix, &nulval,\
      count[0], &anynul, &status ) != 0 ){
    printf( "cannot read from '%s'\n", infile );
    return( status );
  }
  if( fits_close_file( fp, &status ) != 0 ){
    return( status );
  }

  return( 0 );

}

double realcount2( float c[][NX1], int y, double x_low, double x_high ){

  int ix_low, ix_high;
  int i, j;
  double count = 0.0;

  if( y < 1 || y > NY1 || x_low >= x_high || ( x_low < 0.5 && x_high < 0.5 ) || ( x_low > ( NX1 + 0.5 ) && x_high > ( NX1 + 0.5 ) ) ){
    count = 0.0;
  }
  else{
    if( x_low < 0.5 )
      x_low = 0.5;
    if( x_high > ( NX1 + 0.5 ) )
      x_high = NX1 + 0.5;

    ix_low = x_low + 1.4999999999;
    ix_high = x_high - 0.4999999999;

    if( ix_low <= ix_high ){
      for( j=ix_low; j<=ix_high; j++ ){
        count += c[y-1][j-1];
      }
    }
    if( ( ix_high - ix_low ) > -2 ){
      if( (ix_low - 2) >= 0 )
	count += ( ( ( ix_low - 0.5 ) - x_low ) * c[y-1][ix_low-2] );
      if( ix_high <= ( NX1 - 1 ) )
	count += ( ( x_high - ( ix_high + 0.5 ) ) * c[y-1][ix_high] );
    }
    else{
      ix_high = ix_high + 1;
      ix_low = ix_low - 1;
      if( ix_high != ix_low ){
	printf( "Error in realcount\n" );
	return( 0.0 );
      }
      count += ( ( x_high - x_low ) * c[y-1][ix_high-1] );
    }
  }
  return( count );
}

void inversematrix( double a[][LMDORDER], double c[][LMDORDER] ){

  double b[LMDORDER][LMDORDER];
  double coef;
  int ii, i, j, o;

  o = LMDORDER - 1;
  
  for( i=0; i<=o; i++ ){
    for( j=0; j<=o; j++ ){
      b[i][j] = a[i][j];
      if( i == j )
	c[i][j] = 1.0;
      else
	c[i][j] = 0.0;
    }
  }

  for( ii=0; ii<=o; ii++ ){
    coef = b[ii][ii];
    for( j=0; j<=o; j++ ){
      b[ii][j] /= coef;
      c[ii][j] /= coef;
    }
    for( i=0; i<=o; i++ ){
      if( ii != i ){
	coef = b[i][ii];

	for( j=0; j<=o; j++ ){
	  b[i][j] -= ( coef * b[ii][j] );
	  c[i][j] -= ( coef * c[ii][j] );
	}
      }
    }
  }
}

void spmagnifyx( float count1[][NX1], double coef[][FITORDER], double lmd[],\
                 int nline, float count2[][NX1] ){

  int x, y;
  double z;
  double xpeak;
  double beta[LMDORDER];
  double c1[LMDORDER][LMDORDER], c2[LMDORDER][LMDORDER], c3[LMDORDER];
  int i, j, k;
  int fitorder, lmdorder;
  double pos1, pos2;

  fitorder = (int)FITORDER;
  lmdorder = (int)LMDORDER;

  if( fitorder != 4 ){
    printf( "Error: Fitting order in 'hbs1to1x' was not 4.\n" );
    exit( -1 );
  }
  for( y=1; y<=NY1; y++ ){
    for( i=0; i<=(lmdorder-1); i++ ){
      for( j=0; j<=(lmdorder-1); j++ ){
	c1[i][j] = 0.0;
	for( k=0; k<=(nline-1); k++ ){
	  c1[i][j] += ( pow( lmd[k], (double)i ) * pow( lmd[k], (double)j ) );
	}
      }
      c3[i] = 0.0;
      for( k=0; k<=(nline-1); k++ ){
        xpeak = coef[k][0] + coef[k][1]*y + coef[k][2]*y*y + coef[k][3]*y*y*y;
	c3[i] += ( pow( lmd[k], (double)i ) * xpeak );
      }
    }
    inversematrix( c1, c2 );

    for( i=0; i<=(lmdorder-1); i++ ){
      beta[i] = 0.0;
      for( j=0; j<=(lmdorder-1); j++ ){
	beta[i] += ( c2[i][j] * c3[j] );
      }
    }

    for( x=1; x<=NX1; x++ ){
      z = (double)x - 0.50;
      pos1 = ( ( ( ( ( ( beta[7] * z + beta[6] ) * z + beta[5] ) * z\
	     + beta[4] ) * z + beta[3] ) * z + beta[2] ) * z + beta[1] )\
	     *z +beta[0];
      z = (double)x + 0.50;
      pos2 = ( ( ( ( ( ( beta[7] * z + beta[6] ) * z + beta[5] ) * z\
	     + beta[4] ) * z + beta[3] ) * z + beta[2] ) * z + beta[1] )\
	     *z +beta[0];
      count2[y-1][x-1] = realcount2( count1, y, pos1, pos2 );
    }
  }
}

int writefile( char *outfile, float count[][NX1] ){

  fitsfile *fp;
  int i, status;
  char buf[100];
  long naxis1, naxis2, npix;
  char comment[100];

  status = 0;

  if( fits_open_file( &fp, outfile, READWRITE, &status ) != 0 ){
    printf( "Cannot open file '%s' with READWRITE mode. status=%d\n",\
	    outfile, status );
    return( status );
  }
  if( fits_read_key( fp, TLONG, "NAXIS1", &naxis1,\
      comment, &status ) != 0 ){
    return( status );
  }
  if( fits_read_key( fp, TLONG, "NAXIS2", &naxis2,\
      comment, &status ) != 0 ){
    return( status );
  }
  if( naxis1 != NX1 ){
    printf( "Error: definition of NX1 in 'hbsdefine.h'\n" );
    return( -1 );
  }
  if( naxis2 != NY1 ){
    printf( "Error: definition of y-dimension in 'hbsdefine.h'\n" );
    return( -1 );
  }
  npix = naxis1 * naxis2;

  if( fits_write_img( fp, TFLOAT, 1, npix, count, &status ) != 0 ){
    printf( "Error in the function fits_write_img\n" );
    return( status );
  }
  if( fits_close_file( fp, &status ) != 0 ){
    return( status );
  }
  return( 0 );
}

int main( int argc, char *argv[] ){

  double coef[30][FITORDER], lmd[30];
     /* If the number of line of input file exceed 30, correct */
  float count1[NY1][NX1];
  float count2[NY1][NX1];
  int i;
  char infile[50], outfile[50], fn_peakfit[50], fn_lmd[50];
  int n_peak;

  if( argc != 5 ){
    printf( "usage: hbstfx infile fit_table ref_lambda outfile\n" );
    exit( -1 );
  }

  sprintf( infile, "%s", argv[1] );
  strcpy( fn_peakfit, argv[2] );
  strcpy( fn_lmd, argv[3] );
  strcpy( outfile, argv[4] );

  if( n_peak = readpeakfit( fn_peakfit, coef ), n_peak == 0 ){
    printf( "Read error in '%s'\n", fn_peakfit );
    printf( "Do 'hbs1to1y.cl' again !!\n" );
    exit( -1 );
  }
  if( n_peak != readlmd( fn_lmd, lmd ) ){
    printf( "File contains different lines in '%s'\n", fn_lmd );
    printf( "Do 'hbs1to1y.cl' again !!\n" );
    exit( -1 );
  }
  for( i=0; i<=(n_peak-1); i++ ){
    lmd[i] = ( lmd[i] - SW ) / DW + 1.0;/* wavelength --> pixel */
  }

  if( scanfile( infile, count1 ) != 0 ){
    printf( "read error in '%s'\n", infile );
    exit( -1 );
  }

  spmagnifyx( count1, coef, lmd, n_peak, count2 );

  if( writefile( outfile, count2 ) != 0 ){
    printf( "write failed in '%s'\n", outfile );
    return( 1 );
  }

  return( 0 );
}
